/*
* Copyright (c) 2019 Mellanox Technologies, Inc.  All rights reserved.
*
* This software is available to you under a choice of one of two
* licenses.  You may choose to be licensed under the terms of the GNU
* General Public License (GPL) Version 2, available from the file
* COPYING in the main directory of this source tree, or the
* OpenIB.org BSD license below:
*
*     Redistribution and use in source and binary forms, with or
*     without modification, are permitted provided that the following
*     conditions are met:
*
*      - Redistributions of source code must retain the above
*        copyright notice, this list of conditions and the following
*        disclaimer.
*
*      - Redistributions in binary form must reproduce the above
*        copyright notice, this list of conditions and the following
*        disclaimer in the documentation and/or other materials
*        provided with the distribution.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
* BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
* ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/

#include <inttypes.h>
#include <stdarg.h>
#include "mlx5dv_dr.h"

#define HW_STE_SIZE	48
#define JSON_EMPTY_STR	"\"\""

struct dr_dump_ctx {
	struct dr_json_dump *json_file;
	struct mlx5dv_dr_domain *dmn;
};

struct dr_json_dump {
	FILE		*output;
	char		delm;
};

static void dr_json_eor(struct dr_json_dump *dump)
{
	if (dump->delm != '\0')
		putc(dump->delm, dump->output);
	dump->delm = ',';
}

static struct dr_json_dump* dr_json_dump_begin(FILE *f)
{
	struct dr_json_dump *dump = malloc(sizeof(*dump));
	if (dump) {
		dump->output = f;
		dump->delm = '\0';
	}
	return dump;
}

static void dr_json_dump_end(struct dr_json_dump **dump_p)
{
	struct dr_json_dump *dump = *dump_p;

	fputs("\n", dump->output);
	fflush(dump->output);
	free(dump);
	*dump_p = NULL;
}

static void dr_json_key(struct dr_json_dump *dump, const char *name)
{
	dr_json_eor(dump);
	dump->delm = '\0';
	fprintf(dump->output, "\"%s\"", name);
	putc(':', dump->output);
}

static void dr_json_begin(struct dr_json_dump *dump, int c)
{
	dr_json_eor(dump);
	putc(c, dump->output);
	dump->delm = '\0';
}

static void dr_json_end(struct dr_json_dump *dump, int c)
{
	putc(c, dump->output);
	dump->delm = ',';
}

static void dr_json_object_begin(struct dr_json_dump *dump)
{
	dr_json_begin(dump, '{');
}

static void dr_json_object_end(struct dr_json_dump *dump)
{
	dr_json_end(dump, '}');
}

static void dr_json_array_begin(struct dr_json_dump *dump)
{
	dr_json_begin(dump, '[');
}

static void dr_json_array_end(struct dr_json_dump *dump)
{
	dump->delm = '\0';
	dr_json_end(dump, ']');
}

static void dr_json_printf(struct dr_json_dump *dump, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	dr_json_eor(dump);
	vfprintf(dump->output, fmt, ap);
	va_end(ap);
}

static uint64_t dr_dbg_icm_to_idx(uint64_t icm_addr) {
	return (icm_addr >> 6) & 0xffffffff;
}

//todo remove and parse in python
static const char *domain_type_to_str(enum mlx5dv_dr_domain_type domain)
{
	switch (domain) {
	case MLX5DV_DR_DOMAIN_TYPE_NIC_RX:
		return "RX";
		break;
	case MLX5DV_DR_DOMAIN_TYPE_NIC_TX:
		return "TX";
		break;
	case MLX5DV_DR_DOMAIN_TYPE_FDB:
		return "FDB";
		break;
	default:
		return "NONE";
	}
}

//todo remove and parse in python
static const char *tbl_type_to_str(uint32_t type)
{
	switch (type) {
	case FS_FT_NIC_RX:
		return "NIC_RX";
		break;
	case FS_FT_NIC_TX:
		return "NIC_TX";
		break;
	case FS_FT_FDB:
		return "FDB";
		break;
	default:
		return "NONE";
	}
}

static char * dump_hex_print(char * ptr, uint32_t size)
{
	char *hw_ste_dump = NULL;
	int i;

	hw_ste_dump = malloc((2 * size + 1) * sizeof(char));
	if(!hw_ste_dump)
		return NULL;

	for (i = 0; i < size; i++) {
		if (!ptr[i])
			sprintf(&hw_ste_dump[2 * i],"00");
		else
			sprintf(&hw_ste_dump[2 * i],"%.2x", (uint8_t)ptr[i]);
	}
	return hw_ste_dump;
}

static void print_mask(struct dr_dump_ctx *dump_ctx,
		       struct dr_match_param *mask, uint8_t criteria)
{
	char *dump;
	dr_json_key(dump_ctx->json_file, "mask");
	dr_json_object_begin(dump_ctx->json_file);

	dr_json_key(dump_ctx->json_file, "outer");
	if (criteria & DR_MATCHER_CRITERIA_OUTER) {
		dump = dump_hex_print((char *) &mask->outer, sizeof(mask->outer));
		dr_json_printf(dump_ctx->json_file, "\"%s\"",
			       dump ? dump : JSON_EMPTY_STR);
		free(dump);
	} else {
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
	}

	dr_json_key(dump_ctx->json_file, "inner");
	if (criteria & DR_MATCHER_CRITERIA_INNER) {
		dump = dump_hex_print((char *) &mask->inner, sizeof(mask->inner));
		dr_json_printf(dump_ctx->json_file, "\"%s\"",
			       dump ? dump : JSON_EMPTY_STR);
		free(dump);
	} else {
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
	}

	dr_json_key(dump_ctx->json_file, "misc");
	if (criteria & DR_MATCHER_CRITERIA_MISC) {
		dump = dump_hex_print((char *) &mask->misc, sizeof(mask->misc));
		dr_json_printf(dump_ctx->json_file, "\"%s\"",
			       dump ? dump : JSON_EMPTY_STR);
		free(dump);
	} else {
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
	}

	dr_json_key(dump_ctx->json_file, "misc2");
	if (criteria & DR_MATCHER_CRITERIA_MISC2) {
		dump = dump_hex_print((char *) &mask->misc2, sizeof(mask->misc2));
		dr_json_printf(dump_ctx->json_file, "\"%s\"",
			       dump ? dump : JSON_EMPTY_STR);
		free(dump);
	} else {
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
	}

	dr_json_key(dump_ctx->json_file, "misc3");
	if (criteria & DR_MATCHER_CRITERIA_MISC3) {
		dump = dump_hex_print((char *) &mask->misc3, sizeof(mask->misc3));
		dr_json_printf(dump_ctx->json_file, "\"%s\"",
			       dump ? dump : JSON_EMPTY_STR);
		free(dump);
	} else {
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
	}

	dr_json_object_end(dump_ctx->json_file);
}

static void dr_dbg_dump_action(struct dr_dump_ctx *dump_ctx,
			       struct mlx5dv_dr_action *action)
{
	switch (action->action_type) {
	case DR_ACTION_TYP_DROP:
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
		break;
	case DR_ACTION_TYP_FT:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->dest_tbl->devx_obj->object_id);
		break;
	case DR_ACTION_TYP_QP:
		if(action->dest_qp.is_qp)
			dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
				       action->dest_qp.qp->qp_num);
		else
			dr_json_printf(dump_ctx->json_file, "\"0x%x\"", 0);
		break;
	case DR_ACTION_TYP_CTR:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->ctr.devx_obj->object_id +
			       action->ctr.offset);
		break;
	case DR_ACTION_TYP_TAG:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->flow_tag);
		break;
	case DR_ACTION_TYP_MODIFY_HDR:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->rewrite.index);
		break;
	case DR_ACTION_TYP_METER:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->meter.devx_obj->object_id);
		break;
	case DR_ACTION_TYP_VPORT:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->vport.caps->num);
		break;
	case DR_ACTION_TYP_POP_VLAN:
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
		break;
	case DR_ACTION_TYP_PUSH_VLAN:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->push_vlan.vlan_hdr);
		break;
	case DR_ACTION_TYP_TNL_L2_TO_L2:
		dr_json_printf(dump_ctx->json_file, JSON_EMPTY_STR);
		break;
	case DR_ACTION_TYP_TNL_L3_TO_L2:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->rewrite.index);
		break;
	case DR_ACTION_TYP_L2_TO_TNL_L2:
	case DR_ACTION_TYP_L2_TO_TNL_L3:
		dr_json_printf(dump_ctx->json_file, "\"0x%x\"",
			       action->reformat.dvo->object_id);
		break;
	default:
		dr_dbg(dump_ctx->dmn, "Err: Found unsupported action type: %d\n",
			action->action_type);
	}
}

static int dr_dbg_dump_rule_rx_tx_json(struct dr_dump_ctx *dump_ctx,
				       struct mlx5dv_dr_matcher *matcher,
				       struct dr_rule_rx_tx *rx,
				       const char *rx_tx)
{
	struct dr_rule_member *rule_mem;
	char *hw_ste_dump;

	if (!list_empty(&rx->rule_members_list)) {
		dr_json_key(dump_ctx->json_file, rx_tx);
		dr_json_object_begin(dump_ctx->json_file);
		dr_json_key(dump_ctx->json_file, "ste_arr");
		dr_json_array_begin(dump_ctx->json_file);

		list_for_each(&rx->rule_members_list, rule_mem, list) {

			dr_json_object_begin(dump_ctx->json_file);
			dr_json_printf(dump_ctx->json_file, "\"icm_address\" : \"0x%" PRIx64 "\"",
				       dr_dbg_icm_to_idx(dr_ste_get_icm_addr(rule_mem->ste)));
			dr_json_key(dump_ctx->json_file, "hw_ste");
			hw_ste_dump = dump_hex_print((char *)rule_mem->ste->hw_ste,
						     HW_STE_SIZE);
			dr_json_printf(dump_ctx->json_file, "\"%s\"",
				       hw_ste_dump ? hw_ste_dump : JSON_EMPTY_STR);
			free(hw_ste_dump);
			dr_json_object_end(dump_ctx->json_file);
		}
		dr_json_array_end(dump_ctx->json_file);
		dr_json_object_end(dump_ctx->json_file);
	}
	return 0;
}

static int dr_dbg_dump_rule_json(struct dr_dump_ctx *dump_ctx,
				 struct mlx5dv_dr_matcher *matcher,
				 struct mlx5dv_dr_rule *rule)
{
	struct dr_rule_action_member *action_mem;
	struct dr_rule_rx_tx *rx = &rule->rx;
	struct dr_rule_rx_tx *tx = &rule->tx;

	dr_json_object_begin(dump_ctx->json_file);
	dr_json_printf(dump_ctx->json_file, "\"handle\" : \"0x%" PRIx64 "\"", (uint64_t) rule);

	if (rx->nic_matcher) {
		dr_dbg_dump_rule_rx_tx_json(dump_ctx, matcher, rx, "rx");
	}

	if (tx->nic_matcher) {
		dr_dbg_dump_rule_rx_tx_json(dump_ctx, matcher, tx, "tx");
	}
	if (!list_empty(&rule->rule_actions_list)) {
		dr_json_key(dump_ctx->json_file, "actions_types");
		dr_json_array_begin(dump_ctx->json_file);

		list_for_each(&rule->rule_actions_list, action_mem, list) {
			dr_json_printf(dump_ctx->json_file, "%u", action_mem->action->action_type);
		}
		dr_json_array_end(dump_ctx->json_file);

		dr_json_key(dump_ctx->json_file, "actions_values");
		dr_json_array_begin(dump_ctx->json_file);

		list_for_each(&rule->rule_actions_list, action_mem, list) {
			dr_dbg_dump_action(dump_ctx, action_mem->action);
		}
		dr_json_array_end(dump_ctx->json_file);
	}

	dr_json_object_end(dump_ctx->json_file);
	return 0;
}

static int dr_dbg_dump_matcher_json(struct dr_dump_ctx *dump_ctx,
				    struct mlx5dv_dr_matcher *matcher)
{
	struct dr_matcher_rx_tx *rx = &matcher->rx;
	struct dr_matcher_rx_tx *tx = &matcher->tx;
	struct mlx5dv_dr_rule *rule;
	int i = 0;

	dr_json_object_begin(dump_ctx->json_file);
	dr_json_printf(dump_ctx->json_file, "\"handle\" : \"0x%" PRIx64 "\"", (uint64_t) matcher);
	dr_json_printf(dump_ctx->json_file, "\"priority\" : %d", matcher->prio);
	print_mask(dump_ctx, &matcher->mask, matcher->match_criteria);

	if (rx->nic_tbl) {
		dr_json_key(dump_ctx->json_file, "rx");
		dr_json_object_begin(dump_ctx->json_file);
		dr_json_printf(dump_ctx->json_file, "\"builders_num\" : %d", rx->num_of_builders);
		dr_json_printf(dump_ctx->json_file, "\"e_anchor\" : \"0x%" PRIx64 "\"",
			       dr_dbg_icm_to_idx(rx->e_anchor->chunk->icm_addr));
		dr_json_printf(dump_ctx->json_file, "\"s_anchor\" : \"0x%x\"",
			       dr_dbg_icm_to_idx(rx->s_htbl->chunk->icm_addr));

		dr_json_key(dump_ctx->json_file, "builders");
		dr_json_array_begin(dump_ctx->json_file);

		for (i = 0; i < rx->num_of_builders; i++) {
			dr_json_object_begin(dump_ctx->json_file);
			dr_json_printf(dump_ctx->json_file, "\"lu_type\" : \"0x%x\"",
				       rx->ste_builder[i].lu_type);
			dr_json_object_end(dump_ctx->json_file);

		}

		dr_json_array_end(dump_ctx->json_file);
		dr_json_object_end(dump_ctx->json_file);
	}

	if (tx->nic_tbl) {
		dr_json_key(dump_ctx->json_file, "tx");
		dr_json_object_begin(dump_ctx->json_file);
		dr_json_printf(dump_ctx->json_file, "\"builders_num\" : %d", tx->num_of_builders);
		dr_json_printf(dump_ctx->json_file, "\"e_anchor\" : \"0x%" PRIx64 "\"",
			       dr_dbg_icm_to_idx(tx->e_anchor->chunk->icm_addr));
		dr_json_printf(dump_ctx->json_file, "\"s_anchor\" : \"0x%x\"",
			       dr_dbg_icm_to_idx(tx->s_htbl->chunk->icm_addr));

		dr_json_key(dump_ctx->json_file, "builders");
		dr_json_array_begin(dump_ctx->json_file);

		for (i = 0; i < tx->num_of_builders; i++) {
			dr_json_object_begin(dump_ctx->json_file);
			dr_json_printf(dump_ctx->json_file, "\"lu_type\" : \"0x%x\"",
				       tx->ste_builder[i].lu_type);
			dr_json_object_end(dump_ctx->json_file);
		}

		dr_json_array_end(dump_ctx->json_file);
		dr_json_object_end(dump_ctx->json_file);
	}

	dr_json_key(dump_ctx->json_file, "rules");
	dr_json_array_begin(dump_ctx->json_file);

	if (!list_empty(&matcher->rule_list)) {
		list_for_each(&matcher->rule_list, rule, rule_list) {
			dr_dbg_dump_rule_json(dump_ctx, matcher, rule);
		}
	}

	dr_json_array_end(dump_ctx->json_file);
	dr_json_object_end(dump_ctx->json_file);
	return 0;
}

static int dr_dbg_dump_table_json(struct dr_dump_ctx *dump_ctx,
				  struct mlx5dv_dr_table *tbl)
{
	struct dr_table_rx_tx *rx = &tbl->rx;
	struct dr_table_rx_tx *tx = &tbl->tx;
	struct mlx5dv_dr_matcher *matcher;

	dr_json_object_begin(dump_ctx->json_file);
	dr_json_printf(dump_ctx->json_file, "\"handle\" : \"0x%" PRIx64 "\"",
		       (uint64_t) tbl);
	dr_json_printf(dump_ctx->json_file, "\"type\" : \"%s\"",
		       tbl_type_to_str(tbl->table_type));
	dr_json_printf(dump_ctx->json_file, "\"level\" : %d", tbl->level);
	dr_json_printf(dump_ctx->json_file, "\"id\" : %d",
		       tbl->level ? tbl->devx_obj->object_id : -1);

	if (tbl->level) {
		if (rx->nic_dmn) {
			dr_json_key(dump_ctx->json_file, "rx");
			dr_json_object_begin(dump_ctx->json_file);
			dr_json_printf(dump_ctx->json_file, "\"s_anchor\": \"0x%" PRIx64 "\"",
				       dr_dbg_icm_to_idx(rx->s_anchor->chunk->icm_addr));
			dr_json_object_end(dump_ctx->json_file);
		}

		if (tx->nic_dmn) {
			dr_json_key(dump_ctx->json_file, "tx");
			dr_json_object_begin(dump_ctx->json_file);
			dr_json_printf(dump_ctx->json_file, "\"s_anchor\" : \"0x%" PRIx64 "\"",
				       dr_dbg_icm_to_idx(tx->s_anchor->chunk->icm_addr));
			dr_json_object_end(dump_ctx->json_file);
		}

		dr_json_key(dump_ctx->json_file, "matchers");
		dr_json_array_begin(dump_ctx->json_file);

		if (!list_empty(&tbl->matcher_list)) {
			list_for_each(&tbl->matcher_list, matcher, matcher_list) {
				dr_dbg_dump_matcher_json(dump_ctx, matcher);
			}
		}
		dr_json_array_end(dump_ctx->json_file);
	}
	dr_json_object_end(dump_ctx->json_file);
	return 0;
}

static int dr_dbg_dump_domain_json(struct dr_dump_ctx *dump_ctx,
				   struct mlx5dv_dr_domain *dmn)
{
	struct dr_devx_vport_cap *vport_caps;
	struct mlx5dv_dr_table *tbl;
	int num_vports;
	int i;

	pthread_mutex_unlock(&dmn->mutex);
	num_vports = dmn->info.attr.phys_port_cnt;

	dr_json_object_begin(dump_ctx->json_file);
	dr_json_key(dump_ctx->json_file, "domain");
	dr_json_object_begin(dump_ctx->json_file);
	dr_json_printf(dump_ctx->json_file, "\"handle\" : \"0x%" PRIx64 "\"", (uint64_t) dmn);
	dr_json_printf(dump_ctx->json_file, "\"gvmi\" : %d", dmn->info.caps.gvmi);
	dr_json_printf(dump_ctx->json_file, "\"type\" : \"%s\"", domain_type_to_str(dmn->type));
	dr_json_printf(dump_ctx->json_file, "\"num_vports\" : %d", dmn->info.attr.phys_port_cnt);
	dr_json_key(dump_ctx->json_file, "vports");

	/* Query vports */
	dr_json_array_begin(dump_ctx->json_file);
	for (i = 0; i < num_vports; i++) {
		vport_caps = &dmn->info.caps.vports_caps[i];
		dr_json_object_begin(dump_ctx->json_file);
		dr_json_printf(dump_ctx->json_file, "\"vport_num\" : %d", i);
		dr_json_key(dump_ctx->json_file, "icm_addr_rx");
		dr_json_printf(dump_ctx->json_file, "\"0x%" PRIx64 "\"", vport_caps->icm_address_rx);
		dr_json_key(dump_ctx->json_file, "icm_addr_tx");
		dr_json_printf(dump_ctx->json_file, "\"0x%" PRIx64 "\"", vport_caps->icm_address_tx);
		dr_json_object_end(dump_ctx->json_file);
	}
	dr_json_array_end(dump_ctx->json_file);

	dr_json_key(dump_ctx->json_file, "tables");
	dr_json_array_begin(dump_ctx->json_file);

	if (!list_empty(&dmn->tbl_list)) {
		list_for_each(&dmn->tbl_list, tbl, tbl_list) {
			dr_dbg_dump_table_json(dump_ctx, tbl);
		}
	}

	pthread_mutex_unlock(&dmn->mutex);

	dr_json_array_end(dump_ctx->json_file);
	dr_json_object_end(dump_ctx->json_file);
	dr_json_object_end(dump_ctx->json_file);

	return 0;
}

int mlx5dv_dump_dr_domain(FILE *fout, struct mlx5dv_dr_domain *dmn)
{
	struct dr_dump_ctx *dump_ctx;
	uint32_t ret;

	if (!dmn)
		goto exit_params_error;

	if (!fout) {
		dr_dbg(dmn, "Err: Illegal file param!!!\n");
		goto exit_params_error;
	}

	dump_ctx = malloc(sizeof(*dump_ctx));
	if (!dump_ctx) {
		dr_dbg(dmn, "Err: failed to alloc dump_ctx!!!\n");
		errno = ENOMEM;
		return ENOMEM;
	}

	dump_ctx->json_file = dr_json_dump_begin(fout);
	if (!dump_ctx->json_file)
		dr_dbg(dmn, "Err: failed to create dump file!!!\n");

	dump_ctx->dmn = dmn;
	ret = dr_dbg_dump_domain_json(dump_ctx, dmn);

	if (ret)
		dr_dbg(dmn, "Err: failed to dump domain !!!\n");

	dr_json_dump_end(&dump_ctx->json_file);
	free(dump_ctx);
	return 0;

exit_params_error:
	errno = EINVAL;
	return errno;
}
